
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  NWEXT.C
*   DESCRIP  :  FENRIS NetWare Extended Directory Management
*   DATE     :  January 4, 1999
*
*
***************************************************************************/

#include "globals.h"

extern void NWLockVolume(VOLUME *volume);
extern void NWUnlockVolume(VOLUME *volume);

void NWLockExtDirAssign(VOLUME *volume)
{
#if (LINUX_SLEEP)
    if (WaitOnSemaphore(&volume->ExtDirAssignSemaphore) == -EINTR)
       NWFSPrint("lock ext dir assign was interupted\n");
#endif
}

void NWUnlockExtDirAssign(VOLUME *volume)
{
#if (LINUX_SLEEP)
    SignalSemaphore(&volume->ExtDirAssignSemaphore);
#endif
}

void NWLockExtDirBlock(VOLUME *volume)
{
#if (LINUX_SLEEP)
    if (WaitOnSemaphore(&volume->ExtDirBlockSemaphore) == -EINTR)
       NWFSPrint("lock ext dir block was interupted\n");
#endif
}

void NWUnlockExtDirBlock(VOLUME *volume)
{
#if (LINUX_SLEEP)
    SignalSemaphore(&volume->ExtDirBlockSemaphore);
#endif
}

ULONG ReadExtDirectoryRecord(VOLUME *volume, EXTENDED_DIR *dir,
			     ULONG ExtDirNo)
{
    register ULONG ExtDirsPerCluster, hash, cbytes;
    register ULONG Block, Offset;
    register DIR_BLOCK_HASH_LIST *HashTable;
    register DIR_BLOCK_HASH *list;
    ULONG retCode;

    if (!dir)
       return NwInvalidParameter;

    ExtDirsPerCluster = volume->ClusterSize / sizeof(EXTENDED_DIR);
    Block = ExtDirNo / ExtDirsPerCluster;
    Offset = ExtDirNo % ExtDirsPerCluster;

    hash = (Block & (volume->ExtDirBlockHashLimit - 1));
    HashTable = (DIR_BLOCK_HASH_LIST *) volume->ExtDirBlockHash;
    if (!HashTable)
       return NwInvalidParameter;

    NWLockExtDirBlock(volume);
    list = (DIR_BLOCK_HASH *) HashTable[hash].head;
    while (list)
    {
       if (list->BlockNo == Block)
       {
	  cbytes = ReadJournalClusterWithOffset(volume, list->Cluster1,
					 (Offset * sizeof(EXTENDED_DIR)),
					 (BYTE *)dir, sizeof(EXTENDED_DIR),
					 KERNEL_ADDRESS_SPACE, &retCode,
					 DIRECTORY_PRIORITY);
	  if (cbytes == sizeof(EXTENDED_DIR))
	  {
	     NWUnlockExtDirBlock(volume);
	     return 0;
	  }

	  // if the first dir read failed, try the second
	  cbytes = ReadJournalClusterWithOffset(volume, list->Cluster2,
					 (Offset * sizeof(EXTENDED_DIR)),
					 (BYTE *)dir, sizeof(EXTENDED_DIR),
					 KERNEL_ADDRESS_SPACE, &retCode,
					 DIRECTORY_PRIORITY);
	  if (cbytes == sizeof(EXTENDED_DIR))
	  {
	     NWUnlockExtDirBlock(volume);
	     return 0;
	  }
	  NWUnlockExtDirBlock(volume);
	  return NwDiskIoError;
       }
       list = list->next;
    }
    NWUnlockExtDirBlock(volume);
    return NwNoEntry;

}

ULONG WriteExtDirectoryRecord(VOLUME *volume, EXTENDED_DIR *dir,
			      ULONG ExtDirNo)
{
    register ULONG ExtDirsPerCluster, hash, cbytes;
    register ULONG Block, Offset;
    register DIR_BLOCK_HASH_LIST *HashTable;
    register DIR_BLOCK_HASH *list;
    ULONG retCode;

    if (!dir)
       return NwInvalidParameter;

    ExtDirsPerCluster = volume->ClusterSize / sizeof(EXTENDED_DIR);
    Block = ExtDirNo / ExtDirsPerCluster;
    Offset = ExtDirNo % ExtDirsPerCluster;

    hash = (Block & (volume->ExtDirBlockHashLimit - 1));
    HashTable = (DIR_BLOCK_HASH_LIST *) volume->ExtDirBlockHash;
    if (!HashTable)
       return NwInvalidParameter;

    NWLockExtDirBlock(volume);
    list = (DIR_BLOCK_HASH *) HashTable[hash].head;
    while (list)
    {
       if (list->BlockNo == Block)
       {
	  cbytes = WriteJournalClusterWithOffset(volume, list->Cluster1,
					  (Offset * sizeof(EXTENDED_DIR)),
					  (BYTE *)dir, sizeof(EXTENDED_DIR),
					  KERNEL_ADDRESS_SPACE, &retCode,
					  DIRECTORY_PRIORITY);
	  if (cbytes == sizeof(EXTENDED_DIR))
	  {
	     // if first write succeeded, write mirror directory entry
	     cbytes = WriteJournalClusterWithOffset(volume, list->Cluster2,
					  (Offset * sizeof(EXTENDED_DIR)),
					  (BYTE *)dir, sizeof(EXTENDED_DIR),
					  KERNEL_ADDRESS_SPACE, &retCode,
					  DIRECTORY_PRIORITY);
	     if (cbytes == sizeof(EXTENDED_DIR))
	     {
		NWUnlockExtDirBlock(volume);
		return 0;
	     }
	  }
	  NWUnlockExtDirBlock(volume);
	  return NwDiskIoError;
       }
       list = list->next;
    }
    NWUnlockExtDirBlock(volume);
    return NwNoEntry;

}

ULONG ReadExtendedDirectory(VOLUME *volume)
{
    register FAT_ENTRY *FAT1, *FAT2;
    register ULONG retCode;
    register long index1, index2;
    register ULONG DirNo, i, Block = 0;
    register ULONG cluster1, cluster2, DirCount;
    register BYTE *Buffer1, *Buffer2;
    register EXTENDED_DIR *dir;
    extern ULONG HashExtendedDirectoryRecord(VOLUME *volume, EXTENDED_DIR *dir,
				 ULONG Cluster, ULONG DirNo);

#if (MOUNT_VERBOSE)
    NWFSPrint("*** Validating Extended Directory ***\n");
#endif

    Buffer1 = NWFSCacheAlloc(volume->ClusterSize, EXT_WORKSPACE_TAG);
    if (!Buffer1)
       return -3;

    Buffer2 = NWFSCacheAlloc(volume->ClusterSize, EXT_WORKSPACE_TAG);
    if (!Buffer2)
    {
       NWFSFree(Buffer1);
       return -3;
    }

    volume->EndingExtDirCluster1 = 0;
    volume->EndingExtDirCluster2 = 0;
    volume->EndingExtDirIndex = 0;
    volume->EndingExtBlockNo = 0;
    volume->LastValidExtBlockNo = 0;

    DirNo = 0;
    cluster1 = volume->ExtDirectory1;
    cluster2 = volume->ExtDirectory2;

    // check that both copies of the extended directory are present
    if ((cluster1 && !cluster2) || (!cluster1 && cluster2))
    {
       NWFSPrint("nwfs:  extended directory fat chain mismatch\n");
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return -1;
    }

    // extended directory is null, just return
    if (!cluster1)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return 0;
    }

    retCode = CreateExtDirBlockEntry(volume, cluster1, cluster2, Block);
    if (retCode)
    {
       NWFSPrint("nwfs:  could not allocate Dir Block Hash element in ReadExtDirectory\n");
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return -1;
    }

    // read extended directory root 1

    retCode = ReadJournalVolumeCluster(volume, cluster1, Buffer1, volume->ClusterSize,
    					DIRECTORY_PRIORITY);
    if (retCode)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return retCode;
    }

    // read extended directory root 2

    retCode = ReadJournalVolumeCluster(volume, cluster2, Buffer2, volume->ClusterSize,
    					DIRECTORY_PRIORITY);
    if (retCode)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return retCode;
    }

    FAT1 = GetFatEntry(volume, cluster1);
    FAT2 = GetFatEntry(volume, cluster2);
    if (!FAT1 || !FAT2)
    {
       NWFSPrint("nwfs:  error reading fat entry ReadExtDirectory\n");
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return -1;
    }

    index1 = FAT1->FATIndex;
    index2 = FAT2->FATIndex;
    DirCount = volume->ClusterSize / sizeof(DOS);

    while (TRUE)
    {
       if (NWFSCompare(Buffer1, Buffer2, volume->ClusterSize))
       {
	  extern void dumpRecordBytes(BYTE *, ULONG);

	  NWFSPrint("nwfs:  Extended Directory data mirror mismatch\n");

	  NWFSPrint("index1-%d cluster1-[%08X] next1-[%08X]\n", (int)index1,
		    (unsigned int)cluster1,
		    (unsigned int)FAT1->FATCluster);
	  dumpRecordBytes(Buffer1, 64);

	  NWFSPrint("index2-%d cluster2-[%08X] next2-[%08X]\n", (int)index2,
		    (unsigned int)cluster2,
		    (unsigned int)FAT2->FATCluster);
	  dumpRecordBytes(Buffer2, 64);

	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  return -3;
       }

       for (i=0; i < DirCount; i++, DirNo++)
       {
	  dir = (EXTENDED_DIR *) &Buffer1[i * sizeof(ROOT)];
	  retCode = HashExtendedDirectoryRecord(volume, dir, cluster1, DirNo);
	  if (retCode)
	  {
	     NWFSPrint("nwfs:  not enough memory to hash extended directory block\n");
	     NWFSFree(Buffer1);
	     NWFSFree(Buffer2);
	     return retCode;
	  }
       }

       if (FAT1->FATCluster == (ULONG) -1)
       {
	  break;
       }

       // if the other FAT chain terminated abruptly, exit the loop

       if (FAT2->FATCluster == (ULONG) -1)
       {
	  break;
       }

       if (!FAT1->FATCluster)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Free Cluster detected in Ext DIR1 Chain\n");
	  return -3;
       }

       if (!FAT2->FATCluster)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Free Cluster detected in Ext DIR2 Chain\n");
	  return -3;
       }

       if (FAT1->FATCluster & 0x80000000)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  SubAlloc Node Detected in Ext DIR1\n");
	  return -3;
       }

       if (FAT2->FATCluster & 0x80000000)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  SubAlloc Node Detected in Ext DIR2\n");
	  return -3;
       }

       Block++;
       cluster1 = FAT1->FATCluster;
       cluster2 = FAT2->FATCluster;

       retCode = CreateExtDirBlockEntry(volume, cluster1, cluster2, Block);
       if (retCode)
       {
	  NWFSPrint("nwfs:  could not allocate Dir Block Hash element in ReadExtDirectory\n");
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  return retCode;
       }

       retCode = ReadJournalVolumeCluster(volume, cluster1, Buffer1, volume->ClusterSize,
       					   DIRECTORY_PRIORITY);
       if (retCode)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Read Error Ext Directory1 returned %d\n", (int)retCode);
	  return retCode;
       }

       retCode = ReadJournalVolumeCluster(volume, cluster2, Buffer2, volume->ClusterSize,
       				           DIRECTORY_PRIORITY);
       if (retCode)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Read Error Ext Directory2 returned %d\n", (int)retCode);
	  return retCode;
       }

       FAT1 = GetFatEntry(volume, cluster1);
       FAT2 = GetFatEntry(volume, cluster2);
       if (!FAT1 || !FAT2)
       {
	  NWFSPrint("nwfs:  error reading fat entry ReadExtDirectory\n");
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  return -1;
       }

       // index numbers should always be ascending.  if we detect a
       // non-ascending sequence, this indicates a corrupt FAT chain

       if (index1 > (long)FAT1->FATIndex)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  EXTDIR1 fat index overlaps on itself\n");
	  return -1;
       }

       if (index2 > (long)FAT2->FATIndex)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  EXTDIR2 fat index overlaps on itself\n");
	  return -1;
       }

       //  update FAT index number for chains

       index1 = FAT1->FATIndex;
       index2 = FAT2->FATIndex;

    }

    if (FAT1->FATCluster != FAT2->FATCluster)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       NWFSPrint("nwfs:  Extended Directory FAT chain mirror mismatch\n");
       return -3;
    }

    NWFSFree(Buffer1);
    NWFSFree(Buffer2);

    return 0;

}

#if (WINDOWS_NT_UTIL | LINUX_UTIL | DOS_UTIL)

ULONG DumpExtendedDirectoryFile(VOLUME *volume, FILE *fp)
{
    register FAT_ENTRY *FAT1, *FAT2;
    register ULONG retCode;
    register long index1, index2;
    register ULONG i, Block = 0;
    register ULONG cluster1, cluster2;
    register BYTE *Buffer1, *Buffer2;
    register ULONG widget = 0;
#if (LINUX_UTIL)
    struct termios init_settings, new_settings;
#endif
    BYTE wa[4] = "\\|/-";

    NWFSPrint("Analyzing Extended Directory ... ");
    NWFSPrint("%c\b", wa[widget++ & 3]);

    Buffer1 = NWFSCacheAlloc(volume->ClusterSize, EXT_WORKSPACE_TAG);
    if (!Buffer1)
       return -3;

    Buffer2 = NWFSCacheAlloc(volume->ClusterSize, EXT_WORKSPACE_TAG);
    if (!Buffer2)
    {
       NWFSFree(Buffer1);
       return -3;
    }

    cluster1 = volume->ExtDirectory1;
    cluster2 = volume->ExtDirectory2;

    // check that both copies of the extended directory are present
    if ((cluster1 && !cluster2) || (!cluster1 && cluster2))
    {
       NWFSPrint("nwfs:  extended directory fat chain mismatch\n");
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return -1;
    }

    // extended directory is null, just return
    if (!cluster1)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return 0;
    }

    // read extended directory root 1

    retCode = ReadJournalVolumeCluster(volume, cluster1, Buffer1, volume->ClusterSize, DIRECTORY_PRIORITY);
    if (retCode)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return retCode;
    }

    // read extended directory root 2

    retCode = ReadJournalVolumeCluster(volume, cluster2, Buffer2, volume->ClusterSize, DIRECTORY_PRIORITY);
    if (retCode)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return retCode;
    }

    FAT1 = GetFatEntry(volume, cluster1);
    FAT2 = GetFatEntry(volume, cluster2);
    if (!FAT1 || !FAT2)
    {
       NWFSPrint("nwfs:  error reading fat entry ReadExtDirectory\n");
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       return -1;
    }

    index1 = FAT1->FATIndex;
    index2 = FAT2->FATIndex;

#if (LINUX_UTIL)
    // if we are Linux, we need kbhit(), so we implement a terminal
    // based version of kbhit() so we can detect keystrokes during
    // the imaging operation, and abort the operation if asked to
    // do so by the user.  the code below puts a terminal into raw mode
    // so that is does not block when trying to read from the
    // keyboard device.

    tcgetattr(0, &init_settings);
    new_settings = init_settings;
    new_settings.c_lflag &= ~ICANON;
    new_settings.c_lflag &= ~ECHO;
    new_settings.c_lflag &= ~ISIG;
    new_settings.c_cc[VMIN] = 1;
    new_settings.c_cc[VTIME] = 0;
    tcsetattr(0, TCSANOW, &new_settings);
#endif

    while (TRUE)
    {
#if (LINUX_UTIL)
       BYTE nread, ch;

       // put the terminal into raw mode, and see if we have a valid
       // keystroke pending.  if we have a keystroke pending, then
       // see if the user requested that we abort the imaging operation.

       new_settings.c_cc[VMIN] = 0;
       tcsetattr(0, TCSANOW, &new_settings);
       nread = read(0, &ch, 1);
       new_settings.c_cc[VMIN] = 1;
       tcsetattr(0, TCSANOW, &new_settings);

       if (nread)
       {
	  if ((toupper(ch) == 'X') || (toupper(ch) == 'Q') ||
	      (toupper(ch) == ESC) || (toupper(ch) == 'A'))
	  {
	     tcsetattr(0, TCSANOW, &init_settings);
	     return -13;
	  }
       }
#else
       BYTE ch;

       // DOS and Windows NT/2000 both support the kbhit() function.

       if (kbhit())
       {
	  ch = getch();
	  if ((toupper(ch) == 'X') || (toupper(ch) == 'Q') ||
	      (toupper(ch) == ESC) || (toupper(ch) == 'A'))
	  {
	     return -13;
	  }
       }
#endif
       if (NWFSCompare(Buffer1, Buffer2, volume->ClusterSize))
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  extended Directory data mirror mismatch\n");
	  return -3;
       }

       NWFSPrint("%c\b", wa[widget++ & 3]);
       for (i=0; i < (volume->ClusterSize / IO_BLOCK_SIZE); i++)
	  fwrite(&Buffer1[i * IO_BLOCK_SIZE], IO_BLOCK_SIZE, 1, fp);

       if (FAT1->FATCluster == (ULONG) -1)
       {
	  break;
       }

       // if the other FAT chain terminated abruptly, exit the loop

       if (FAT2->FATCluster == (ULONG) -1)
       {
	  break;
       }

       if (!FAT1->FATCluster)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Free Cluster detected in Ext DIR1 Chain\n");
	  return -3;
       }

       if (!FAT2->FATCluster)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Free Cluster detected in Ext DIR2 Chain\n");
	  return -3;
       }

       if (FAT1->FATCluster & 0x80000000)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  SubAlloc Node Detected in Ext DIR1\n");
	  return -3;
       }

       if (FAT2->FATCluster & 0x80000000)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  SubAlloc Node Detected in Ext DIR2\n");
	  return -3;
       }

       Block++;
       cluster1 = FAT1->FATCluster;
       cluster2 = FAT2->FATCluster;


       retCode = ReadJournalVolumeCluster(volume, cluster1, Buffer1, volume->ClusterSize, DIRECTORY_PRIORITY);
       if (retCode)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Read Error Ext Directory1 returned %d\n", (int)retCode);
	  return retCode;
       }

       retCode = ReadJournalVolumeCluster(volume, cluster2, Buffer2, volume->ClusterSize, DIRECTORY_PRIORITY);
       if (retCode)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  Read Error Ext Directory2 returned %d\n", (int)retCode);
	  return retCode;
       }

       FAT1 = GetFatEntry(volume, cluster1);
       FAT2 = GetFatEntry(volume, cluster2);
       if (!FAT1 || !FAT2)
       {
	  NWFSPrint("nwfs:  error reading fat entry ReadExtDirectory\n");
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  return -1;
       }

       // index numbers should always be ascending.  if we detect a
       // non-ascending sequence, this indicates a corrupt FAT chain

       if (index1 > FAT1->FATIndex)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  EXTDIR1 fat index overlaps on itself\n");
	  return -1;
       }

       if (index2 > FAT2->FATIndex)
       {
	  NWFSFree(Buffer1);
	  NWFSFree(Buffer2);
	  NWFSPrint("nwfs:  EXTDIR2 fat index overlaps on itself\n");
	  return -1;
       }

       //  update FAT index number for chains

       index1 = FAT1->FATIndex;
       index2 = FAT2->FATIndex;

    }
#if (LINUX_UTIL)
    tcsetattr(0, TCSANOW, &init_settings);
#endif

    if (FAT1->FATCluster != FAT2->FATCluster)
    {
       NWFSFree(Buffer1);
       NWFSFree(Buffer2);
       NWFSPrint("nwfs:  Extended Directory FAT chain mirror mismatch\n");
       return -3;
    }

    NWFSFree(Buffer1);
    NWFSFree(Buffer2);

    return 0;

}

#endif


ULONG InitializeExtDirBlockHash(VOLUME *volume)
{
    volume->ExtDirListLock = volume->ExtDirListBlocks = 0;
    volume->ExtDirBlockHash = NWFSCacheAlloc(BLOCK_NUMBER_HASH_SIZE,
					     EXT_BLOCKHASH_TAG);

    if (!volume->ExtDirBlockHash)
       return -1;

    volume->ExtDirBlockHashLimit = NUMBER_OF_BLOCK_HASH_ENTRIES;
    NWFSSet(volume->ExtDirBlockHash, 0, BLOCK_NUMBER_HASH_SIZE);

    return 0;
}

ULONG CreateExtDirBlockEntry(VOLUME *volume, ULONG cluster,
			 ULONG mirror, ULONG index)
{
    register DIR_BLOCK_HASH *hash;
    register ULONG retCode;

    hash = (DIR_BLOCK_HASH *) NWFSAlloc(sizeof(DIR_BLOCK_HASH),
					EXT_BLOCK_TAG);
    if (hash)
    {
       NWFSSet(hash, 0, sizeof(DIR_BLOCK_HASH));

       hash->Cluster1 = cluster;
       hash->Cluster2 = mirror;
       hash->BlockNo = index;

       retCode = AddToExtDirBlockHash(volume, hash);
       if (retCode)
       {
	  NWFSFree(hash);
	  return -1;
       }

       retCode = SetAssignedClusterValue(volume, cluster, 1);
       if (retCode)
       {
	  NWFSPrint("nwfs:  assigned bit block value not set cluster-%X\n",
		    (unsigned int)cluster);
       }

       retCode = SetAssignedClusterValue(volume, mirror, 1);
       if (retCode)
       {
	  NWFSPrint("nwfs:  assigned bit block value not set mirror-%X\n",
		    (unsigned int)mirror);
       }

       return 0;
    }
    return -1;

}

ULONG FreeExtDirBlockHash(VOLUME *volume)
{
    register DIR_BLOCK_HASH_LIST *HashTable;
    register DIR_BLOCK_HASH *list, *hash;
    register ULONG count;

    if (volume->ExtDirBlockHash)
    {
       NWLockExtDirBlock(volume);
       HashTable = (DIR_BLOCK_HASH_LIST *) volume->ExtDirBlockHash;
       for (count=0; count < volume->ExtDirBlockHashLimit; count++)
       {
	  list = (DIR_BLOCK_HASH *) HashTable[count].head;
	  HashTable[count].head = HashTable[count].head = 0;
	  while (list)
	  {
	     hash = list;
	     list = list->next;
	     NWFSFree(hash);
	  }
       }
       NWUnlockExtDirBlock(volume);
    }

    if (volume->ExtDirBlockHash)
       NWFSFree(volume->ExtDirBlockHash);
    volume->ExtDirBlockHashLimit =0;
    volume->ExtDirBlockHash = 0;

    return 0;

}

ULONG AddToExtDirBlockHash(VOLUME *volume, DIR_BLOCK_HASH *dblock)
{
    register ULONG hash;
    register DIR_BLOCK_HASH_LIST *HashTable;

    hash = (dblock->BlockNo & (volume->ExtDirBlockHashLimit - 1));
    HashTable = (DIR_BLOCK_HASH_LIST *) volume->ExtDirBlockHash;
    if (HashTable)
    {
       NWLockExtDirBlock(volume);
       if (!HashTable[hash].head)
       {
	  HashTable[hash].head = dblock;
	  HashTable[hash].tail = dblock;
	  dblock->next = dblock->prior = 0;
       }
       else
       {
	  HashTable[hash].tail->next = dblock;
	  dblock->next = 0;
	  dblock->prior = HashTable[hash].tail;
	  HashTable[hash].tail = dblock;
       }
       NWUnlockExtDirBlock(volume);
       return 0;
    }
    return -1;
}

ULONG RemoveExtDirBlockHash(VOLUME *volume, DIR_BLOCK_HASH *dblock)
{
    register ULONG hash;
    register DIR_BLOCK_HASH_LIST *HashTable;

    hash = (dblock->BlockNo & (volume->ExtDirBlockHashLimit - 1));
    HashTable = (DIR_BLOCK_HASH_LIST *) volume->ExtDirBlockHash;
    if (HashTable)
    {
       NWLockExtDirBlock(volume);
       if (HashTable[hash].head == dblock)
       {
	  HashTable[hash].head = dblock->next;
	  if (HashTable[hash].head)
	     HashTable[hash].head->prior = NULL;
	  else
	     HashTable[hash].tail = NULL;
       }
       else
       {
	  dblock->prior->next = dblock->next;
	  if (dblock != HashTable[hash].tail)
	     dblock->next->prior = dblock->prior;
	  else
	     HashTable[hash].tail = dblock->prior;
       }
       NWUnlockExtDirBlock(volume);
       return 0;
    }
    return -1;
}

//
//
//

ULONG InitializeExtDirAssignHash(VOLUME *volume)
{
    volume->ExtDirAssignLock = volume->ExtDirAssignBlocks = 0;
    volume->ExtDirAssignHash = NWFSCacheAlloc(ASSIGN_BLOCK_HASH_SIZE,
					      ASSN_BLOCKHASH_TAG);

    if (!volume->ExtDirAssignHash)
       return -1;

    volume->ExtDirAssignHashLimit = NUMBER_OF_ASSIGN_BLOCK_ENTRIES;
    NWFSSet(volume->ExtDirAssignHash, 0, ASSIGN_BLOCK_HASH_SIZE);

    return 0;
}

ULONG CreateExtDirAssignEntry(VOLUME *volume, ULONG Parent, ULONG BlockNo,
			      DOS *dos)
{
    register ULONG ExtDirsPerBlock;
    register DIR_ASSIGN_HASH *hash;
    register ULONG retCode, i;

    ExtDirsPerBlock = volume->BlockSize / sizeof(ROOT);
    hash = (DIR_ASSIGN_HASH *) NWFSAlloc(sizeof(DIR_ASSIGN_HASH) +
					 (ExtDirsPerBlock / 8), ASSN_BLOCK_TAG);
    if (hash)
    {
       NWFSSet(hash, 0, sizeof(DIR_ASSIGN_HASH));

       hash->DirOwner = Parent;
       hash->BlockNo = BlockNo;

       // create directory record free list for this block

       for (i=0; i < ExtDirsPerBlock; i++)
       {
	  if (dos[i].Subdirectory == (ULONG) -1)
	     hash->FreeList[i >> 3] &= ~(1 << (i & 7));
	  else
	     hash->FreeList[i >> 3] |= (1 << (i & 7));
       }

       retCode = AddToExtDirAssignHash(volume, hash);
       if (retCode)
       {
	  NWFSFree(hash);
	  return -1;
       }
       return 0;
    }
    return -1;

}

ULONG FreeExtDirAssignHash(VOLUME *volume)
{
    register DIR_ASSIGN_HASH_LIST *HashTable;
    register DIR_ASSIGN_HASH *list, *hash;
    register ULONG count;

    if (volume->ExtDirAssignHash)
    {
       NWLockExtDirAssign(volume);
       HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
       for (count=0; count < volume->ExtDirAssignHashLimit; count++)
       {
	  list = (DIR_ASSIGN_HASH *) HashTable[count].head;
	  HashTable[count].head = HashTable[count].head = 0;
	  while (list)
	  {
	     hash = list;
	     list = list->next;
	     NWFSFree(hash);
	  }
       }
       NWUnlockExtDirAssign(volume);
    }

    if (volume->ExtDirAssignHash)
       NWFSFree(volume->ExtDirAssignHash);
    volume->ExtDirAssignHashLimit = 0;
    volume->ExtDirAssignHash = 0;

    return 0;

}

ULONG AddToExtDirAssignHash(VOLUME *volume, DIR_ASSIGN_HASH *dblock)
{
    register ULONG hash;
    register DIR_ASSIGN_HASH_LIST *HashTable;

    hash = (dblock->DirOwner & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (HashTable)
    {
       NWLockExtDirAssign(volume);
       if (!HashTable[hash].head)
       {
	  HashTable[hash].head = dblock;
	  HashTable[hash].tail = dblock;
	  dblock->next = dblock->prior = 0;
       }
       else
       {
	  HashTable[hash].tail->next = dblock;
	  dblock->next = 0;
	  dblock->prior = HashTable[hash].tail;
	  HashTable[hash].tail = dblock;
       }
       NWUnlockExtDirAssign(volume);
       return 0;
    }
    return -1;
}

ULONG RemoveExtDirAssignHash(VOLUME *volume, DIR_ASSIGN_HASH *dblock)
{
    register ULONG hash;
    register DIR_ASSIGN_HASH_LIST *HashTable;

    hash = (dblock->DirOwner & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (HashTable)
    {
       NWLockExtDirAssign(volume);
       if (HashTable[hash].head == dblock)
       {
	  HashTable[hash].head = dblock->next;
	  if (HashTable[hash].head)
	     HashTable[hash].head->prior = NULL;
	  else
	     HashTable[hash].tail = NULL;
       }
       else
       {
	  dblock->prior->next = dblock->next;
	  if (dblock != HashTable[hash].tail)
	     dblock->next->prior = dblock->prior;
	  else
	     HashTable[hash].tail = dblock->prior;
       }
       NWUnlockExtDirAssign(volume);
       return 0;
    }
    return -1;
}

ULONG AddToExtDirAssignHashNoLock(VOLUME *volume, DIR_ASSIGN_HASH *dblock)
{
    register ULONG hash;
    register DIR_ASSIGN_HASH_LIST *HashTable;

    hash = (dblock->DirOwner & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (HashTable)
    {
       if (!HashTable[hash].head)
       {
	  HashTable[hash].head = dblock;
	  HashTable[hash].tail = dblock;
	  dblock->next = dblock->prior = 0;
       }
       else
       {
	  HashTable[hash].tail->next = dblock;
	  dblock->next = 0;
	  dblock->prior = HashTable[hash].tail;
	  HashTable[hash].tail = dblock;
       }
       return 0;
    }
    return -1;
}

ULONG RemoveExtDirAssignHashNoLock(VOLUME *volume, DIR_ASSIGN_HASH *dblock)
{
    register ULONG hash;
    register DIR_ASSIGN_HASH_LIST *HashTable;

    hash = (dblock->DirOwner & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (HashTable)
    {
       if (HashTable[hash].head == dblock)
       {
	  HashTable[hash].head = dblock->next;
	  if (HashTable[hash].head)
	     HashTable[hash].head->prior = NULL;
	  else
	     HashTable[hash].tail = NULL;
       }
       else
       {
	  dblock->prior->next = dblock->next;
	  if (dblock != HashTable[hash].tail)
	     dblock->next->prior = dblock->prior;
	  else
	     HashTable[hash].tail = dblock->prior;
       }
       return 0;
    }
    return -1;
}

ULONG AllocateExtDirectoryRecord(VOLUME *volume, ULONG Parent)
{
    register ULONG hash, i, ExtDirsPerBlock, DirNo, ExtDirsPerCluster;
    register DIR_ASSIGN_HASH_LIST *HashTable;
    register DIR_ASSIGN_HASH *list;
    register VOLUME_WORKSPACE *WorkSpace;
    register BYTE *cBuffer;
    register ULONG cbytes, vindex, retCode, block, BlockNo;
    register long cluster1, cluster2;
    register DOS *dos;
    ULONG retRCode;

    ExtDirsPerBlock = volume->BlockSize / sizeof(EXTENDED_DIR);
    hash = (Parent & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (!HashTable)
       return -1;

    NWLockExtDirAssign(volume);
    list = (DIR_ASSIGN_HASH *) HashTable[hash].head;
    while (list)
    {
       if (list->DirOwner == Parent)
       {
	  for (i=0; i < ExtDirsPerBlock; i++)
	  {
	     // if we locate a free entry in the bit list, set it
	     // then return the allocated DirNo
	     if (!(((list->FreeList[i >> 3]) >> (i & 7)) & 1))
	     {
		list->FreeList[i >> 3] |= (1 << (i & 7));
		DirNo = ((ExtDirsPerBlock * list->BlockNo) + i);

		NWUnlockExtDirAssign(volume);
		return DirNo;
	     }
	  }

       }
       list = list->next;
    }

    // if we got here, then we could not locate an assigned directory
    // block with available entries.  at this point, we scan the free
    // list hash (-1) and search free directory blocks for available
    // entries.  if we find a free block, then we assign this
    // dir block as owned by the specified parent.

    hash = ((ULONG)FREE_NODE & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (!HashTable)
    {
       NWUnlockExtDirAssign(volume);
       return -1;
    }

    list = (DIR_ASSIGN_HASH *) HashTable[hash].head;
    while (list)
    {
       if (list->DirOwner == FREE_NODE)
       {
	  for (i=0; i < ExtDirsPerBlock; i++)
	  {
	     // if we locate a free entry in the bit list, set it
	     // then return the allocated DirNo
	     if (!(((list->FreeList[i >> 3]) >> (i & 7)) & 1))
	     {
		// set directory entry as allocated
		list->FreeList[i >> 3] |= (1 << (i & 7));

		// re-hash block assign record to reflect
		// new parent
		RemoveExtDirAssignHashNoLock(volume, list);
		list->DirOwner = Parent;
		AddToExtDirAssignHashNoLock(volume, list);

		// calculate and return directory number
		DirNo = ((ExtDirsPerBlock * list->BlockNo) + i);

		NWUnlockExtDirAssign(volume);
		return DirNo;
	     }
	  }
       }
       list = list->next;
    }
    NWUnlockExtDirAssign(volume);

    // if we got here, then the directory is completely full, and we need
    // to extend the directory by allocating another cluster for both the
    // primary and mirror copies of the directory.

    WorkSpace = AllocateWorkspace(volume);
    if (!WorkSpace)
       return -1;
    cBuffer = &WorkSpace->Buffer[0];

    // initialize new cluster and fill with free directory
    // entries
    NWFSSet(cBuffer, 0, volume->ClusterSize);
    ExtDirsPerCluster = volume->ClusterSize / sizeof(EXTENDED_DIR);
    for (i=0; i < ExtDirsPerCluster; i++)
    {
       dos = (DOS *) &cBuffer[i * sizeof(EXTENDED_DIR)];
       dos->Subdirectory = FREE_NODE;
    }

    NWLockVolume(volume);

    // set vindex to next cluster index
    vindex = volume->EndingExtDirIndex + 1;

    // allocate cluster for primary directory. we try to get both clusters
    // first before we start modifying the dir fat chain just in case we
    // run out of disk space. this makes error handling and unwinding of
    // a failed create much easier.

    cluster1 = AllocateClusterSetIndexSetChain(volume, vindex, -1);
    if (cluster1 == -1)
    {
       FreeWorkspace(volume, WorkSpace);
       return -1;
    }

    //  allocate cluster for mirrored directory
    cluster2 = AllocateClusterSetIndexSetChain(volume, vindex, -1);
    if (cluster2 == -1)
    {
       FreeCluster(volume, cluster1);  // free the cluster
       FreeWorkspace(volume, WorkSpace);
       return -1;
    }

    // now write both newly allocated clusters with a pre-initialized
    // cluster of free directory records.  if either write fails, then
    // free both clusters.  we have not spliced them into the dir fat
    // chain as of yet.

    cbytes = WriteJournalClusterWithOffset(volume, cluster1, 0, cBuffer,
				    volume->ClusterSize, KERNEL_ADDRESS_SPACE,
				    &retRCode, DIRECTORY_PRIORITY);
    if (cbytes != volume->ClusterSize)
    {
       FreeWorkspace(volume, WorkSpace);
       FreeCluster(volume, cluster1);
       FreeCluster(volume, cluster2);
       return -1;
    }

    cbytes = WriteJournalClusterWithOffset(volume, cluster2, 0, cBuffer,
				    volume->ClusterSize, KERNEL_ADDRESS_SPACE,
				    &retRCode, DIRECTORY_PRIORITY);
    if (cbytes != volume->ClusterSize)
    {
       FreeWorkspace(volume, WorkSpace);
       FreeCluster(volume, cluster1);
       FreeCluster(volume, cluster2);
       return -1;
    }

    // success, now we can link both the primary and mirrored
    // directory fat chains to point to these new clusters.
    SetClusterValue(volume, volume->EndingExtDirCluster1, cluster1);
    SetClusterValue(volume, volume->EndingExtDirCluster2, cluster2);

    // update ending cluster and ending index values
    volume->EndingExtDirCluster1 = cluster1;
    volume->EndingExtDirCluster2 = cluster2;
    volume->EndingExtDirIndex = vindex;

    // at this point, we have successfully extended both the primary and
    // mirrored directories.  now we update in memory meta-data and hash
    // structures to reflect the changes to the directory file.

    retCode = CreateExtDirBlockEntry(volume, cluster1, cluster2, vindex);
    if (retCode)
    {
       NWFSPrint("nwfs:  could not allocate dir block during dir extend\n");
       FreeWorkspace(volume, WorkSpace);
       return -1;
    }

    // set to next logical block number
    BlockNo = volume->EndingExtBlockNo + 1;
    for (block=0; block < volume->BlocksPerCluster; block++, BlockNo++)
    {
       dos = (DOS *) &cBuffer[block * volume->BlockSize];
       retCode = CreateExtDirAssignEntry(volume, FREE_NODE, BlockNo, dos);
       if (retCode)
       {
	  NWFSPrint("nwfs:  could not allocate dir block Assign element during extend\n");
	  FreeWorkspace(volume, WorkSpace);
	  return -1;
       }
       volume->EndingExtBlockNo = BlockNo;  // save new ending block no
    }

    NWUnlockVolume(volume);

    // free the workspace
    FreeWorkspace(volume, WorkSpace);

    // now try to scan the free list hash (-1) and search free
    // directory blocks for available entries.  if we find a free
    // block, then we mark this dir block as owned by the specified
    // parent.  we should never get here unless we successfully
    // extended the directory file.

    hash = ((ULONG)FREE_NODE & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (!HashTable)
       return -1;

    NWLockExtDirAssign(volume);
    list = (DIR_ASSIGN_HASH *) HashTable[hash].head;
    while (list)
    {
       if (list->DirOwner == FREE_NODE)
       {
	  for (i=0; i < ExtDirsPerBlock; i++)
	  {
	     // if we locate a free entry in the bit list, set it
	     // then return the allocated DirNo
	     if (!(((list->FreeList[i >> 3]) >> (i & 7)) & 1))
	     {
		// set directory entry as allocated
		list->FreeList[i >> 3] |= (1 << (i & 7));

		// re-hash block assign record to reflect
		// new parent
		RemoveExtDirAssignHashNoLock(volume, list);
		list->DirOwner = Parent;
		AddToExtDirAssignHashNoLock(volume, list);

		// calculate and return directory number
		DirNo = ((ExtDirsPerBlock * list->BlockNo) + i);

		NWUnlockExtDirAssign(volume);
		return DirNo;
	     }
	  }
       }
       list = list->next;
    }

    NWUnlockExtDirAssign(volume);
    return -1;

}

ULONG FreeExtDirectoryRecord(VOLUME *volume, EXTENDED_DIR *dir, ULONG DirNo)
{
    register ULONG hash, i, count;
    register ULONG ExtDirsPerBlock, ExtDirsPerCluster;
    register ULONG DirIndex, DirBlockNo;
    register ULONG retCode;
    register DIR_ASSIGN_HASH_LIST *HashTable;
    register DIR_ASSIGN_HASH *list;

    ExtDirsPerCluster = volume->ClusterSize / sizeof(EXTENDED_DIR);
    ExtDirsPerBlock = volume->BlockSize / sizeof(EXTENDED_DIR);
    DirBlockNo = DirNo / ExtDirsPerBlock;
    DirIndex = DirNo % ExtDirsPerBlock;

    hash = ((ULONG)dir->DirectoryNumber & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (!HashTable)
       return -1;

    NWLockExtDirAssign(volume);
    list = (DIR_ASSIGN_HASH *) HashTable[hash].head;
    while (list)
    {
       if ((list->BlockNo == DirBlockNo) &&
	   (list->DirOwner == dir->DirectoryNumber))
       {
	  // free directory record
	  list->FreeList[DirIndex >> 3] &= ~(1 << (DirIndex & 7));

	  // see if this block has any allocated dir entries, if so
	  // then exit, otherwise, re-hash the block into the
	  // free list

	  for (i=0; i < ExtDirsPerBlock; i++)
	  {
	     // check for allocated entries
	     if (((list->FreeList[i >> 3]) >> (i & 7)) & 1)
	     {
		retCode = WriteExtDirectoryRecord(volume, dir, DirNo);
		if (retCode)
		{
		   NWUnlockExtDirAssign(volume);
		   NWFSPrint("nwfs:  error in directory write\n");
		   return -1;
		}
		NWUnlockExtDirAssign(volume);
		return 0;
	     }
	  }

	  // if we get here the we assume that the entire directory
	  // block is now free.  we remove this block from the
	  // previous assigned parent hash, then re-hash it into the
	  // directory free list (-1)

	  RemoveExtDirAssignHashNoLock(volume, list);
	  list->DirOwner = FREE_NODE;
	  AddToExtDirAssignHashNoLock(volume, list);

	  retCode = WriteExtDirectoryRecord(volume, dir, DirNo);
	  if (retCode)
	  {
	     NWUnlockExtDirAssign(volume);
	     NWFSPrint("nwfs:  error in directory write\n");
	     return -1;
	  }
	  NWUnlockExtDirAssign(volume);
	  return 0;
       }
       list = list->next;
    }

    // if we could not locate the directory entry, then perform
    // a brute force search of the record in all of the parent
    // assign hash records.  Netware does allow parent records
    // to be mixed within a single block in rare circumstances.
    // this should rarely happen, however, and if it ever does
    // it indicates possible directory corruption.

    for (count=0; count < volume->ExtDirAssignHashLimit; count++)
    {
       list = (DIR_ASSIGN_HASH *) HashTable[count].head;
       while (list)
       {
	  if (list->BlockNo == DirBlockNo)
	  {
	     NWFSPrint("nwfs:  ext record freed in non-assigned parent block [%X/%X]\n",
		       (unsigned int)dir->DirectoryNumber,
		       (unsigned int)list->DirOwner);

	     // free directory record
	     list->FreeList[DirIndex >> 3] &= ~(1 << (DirIndex & 7));

	     // see if this block has any allocated dir entries, if so
	     // then exit, otherwise, re-hash the block into the
	     // free list

	     for (i=0; i < ExtDirsPerBlock; i++)
	     {
		// check for allocated entries
		if (((list->FreeList[i >> 3]) >> (i & 7)) & 1)
		{
		   // if we find an allocated record, then clear the
		   // individual directory record and exit

		   retCode = WriteExtDirectoryRecord(volume, dir, DirNo);
		   if (retCode)
		   {
		      NWUnlockExtDirAssign(volume);
		      NWFSPrint("nwfs:  error in directory write\n");
		      return retCode;
		   }
		   NWUnlockExtDirAssign(volume);
		   return 0;
		}
	     }

	     // if we get here the we assume that the entire directory
	     // block is now free.  we remove this block from the
	     // previous assigned parent hash, then re-hash it into the
	     // directory free list (-1)

	     RemoveExtDirAssignHashNoLock(volume, list);
	     list->DirOwner = FREE_NODE;
	     AddToExtDirAssignHashNoLock(volume, list);

	     retCode = WriteExtDirectoryRecord(volume, dir, DirNo);
	     if (retCode)
	     {
		NWUnlockExtDirAssign(volume);
		NWFSPrint("nwfs:  error in directory write\n");
		return retCode;
	     }
	     NWUnlockExtDirAssign(volume);
	     return 0;
	  }
	  list = list->next;
       }
    }
    NWUnlockExtDirAssign(volume);
    return -1;
}

ULONG ReleaseExtDirectoryRecord(VOLUME *volume, ULONG DirNo, ULONG Parent)
{
    register ULONG hash, i, count;
    register ULONG ExtDirsPerBlock, ExtDirsPerCluster;
    register ULONG DirIndex, DirBlockNo;
    register DIR_ASSIGN_HASH_LIST *HashTable;
    register DIR_ASSIGN_HASH *list;

    ExtDirsPerCluster = volume->ClusterSize / sizeof(EXTENDED_DIR);
    ExtDirsPerBlock = volume->BlockSize / sizeof(EXTENDED_DIR);
    DirBlockNo = DirNo / ExtDirsPerBlock;
    DirIndex = DirNo % ExtDirsPerBlock;

    hash = ((ULONG)Parent & (volume->ExtDirAssignHashLimit - 1));
    HashTable = (DIR_ASSIGN_HASH_LIST *) volume->ExtDirAssignHash;
    if (!HashTable)
       return -1;

    NWLockExtDirAssign(volume);
    list = (DIR_ASSIGN_HASH *) HashTable[hash].head;
    while (list)
    {
       if ((list->BlockNo == DirBlockNo) &&
	   (list->DirOwner == Parent))
       {
	  // free directory record
	  list->FreeList[DirIndex >> 3] &= ~(1 << (DirIndex & 7));

	  // see if this block has any allocated dir entries, if so
	  // then exit, otherwise, re-hash the block into the
	  // free list

	  for (i=0; i < ExtDirsPerBlock; i++)
	  {
	     // check for allocated entries
	     if (((list->FreeList[i >> 3]) >> (i & 7)) & 1)
	     {
		NWUnlockExtDirAssign(volume);
		return 0;
	     }
	  }

	  RemoveExtDirAssignHashNoLock(volume, list);
	  list->DirOwner = FREE_NODE;
	  AddToExtDirAssignHashNoLock(volume, list);

	  NWUnlockExtDirAssign(volume);
	  return 0;
       }
       list = list->next;
    }

    // if we could not locate the directory entry, then perform
    // a brute force search of the record in all of the parent
    // assign hash records.  Netware does allow parent records
    // to be mixed within a single block in rare circumstances.
    // this should rarely happen, however, and if it ever does
    // it indicates possible directory corruption.

    for (count=0; count < volume->ExtDirAssignHashLimit; count++)
    {
       list = (DIR_ASSIGN_HASH *) HashTable[count].head;
       while (list)
       {
	  if (list->BlockNo == DirBlockNo)
	  {
	     NWFSPrint("nwfs:  ext record freed in non-assigned parent block [%X/%X]\n",
		       (unsigned int)Parent,
		       (unsigned int)list->DirOwner);

	     // free directory record
	     list->FreeList[DirIndex >> 3] &= ~(1 << (DirIndex & 7));

	     // see if this block has any allocated dir entries, if so
	     // then exit, otherwise, re-hash the block into the
	     // free list

	     for (i=0; i < ExtDirsPerBlock; i++)
	     {
		// check for allocated entries
		// if we find an allocated record, then exit
		if (((list->FreeList[i >> 3]) >> (i & 7)) & 1)
		{
		   NWUnlockExtDirAssign(volume);
		   return 0;
		}
	     }

	     // if we get here the we assume that the entire directory
	     // block is now free.  we remove this block from the
	     // previous assigned parent hash, then re-hash it into the
	     // directory free list (-1)

	     RemoveExtDirAssignHashNoLock(volume, list);
	     list->DirOwner = FREE_NODE;
	     AddToExtDirAssignHashNoLock(volume, list);

	     NWUnlockExtDirAssign(volume);
	     return 0;
	  }
	  list = list->next;
       }
    }
    NWUnlockExtDirAssign(volume);
    return -1;
}



